//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.2
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Version 0.1: 1999/05/05
// Copyright (C) 1999 Jun-ichi Yoshida and Ikki Ohmukai, All rights reserved.
// Version 1.0: 2001/12/13
// Copyright (C) 2001 Jun-ichi Yoshida, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
// Version 1.1: 2002/01/17
// Copyright (C) 2002 Jun-ichi Yoshida, Masaki Sano, Tomoyuki Hiroyasu and Mitsunori Miki, All rights reserved.
// Version 1.2: 2002/06/04
// Copyright (C) 2002 Jun-ichi Yoshida, Masaki Sano, Tomoyuki Hiroyasu and Mitsunori Miki, All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notce and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright
//    notice and this list of conditions in the documentation and/or other
//    materials provided with the distribution.
// 3. The names of copyrighters of this software  should not be used to
//    endorse or promote products derived from this software without specific
//    prior written permission.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ga2k.cpp
//-----------------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <cfloat>
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <numeric>
#include <cmath>
using namespace std;


#include "ga2k.h"
#include "randring.h"

//gray coding
static void gray_code(vector<char>& bitString, int startPoint, int bitLength, int num)
{
	int mask;
	bool f = false;
	for(mask = 1 << (sizeof(int) * 8 - 2); mask != 0; mask >>= 1)
	{
		if(f)
		{
			num ^= mask;
			if(num & mask) f = false;
		}
		else
			if(num & mask) f = true;
	}

	int i = startPoint + bitLength;
	while(startPoint < i){
		bitString[--i] = num & 1;
		num >>= 1;
	}
}


//gray decode
static int gray_decode(const vector<char>& bitString, int startPoint, int bitLength)
{
	int i, real, mask;
	bool sw = false;

	for(i = real = 0; i < bitLength; i++)
	{
		if(bitString[startPoint + i])
			real |= 1 << (bitLength - i - 1);
	}

	for(mask = 1 << (bitLength - 1); mask != 0; mask >>= 1)
	{
		if(sw)
		{
			if(real & mask) sw = false;
			real ^= mask;
		}
		else if(real & mask) sw = true;
	}
	return real;
}

//binary decode
static int binary_decode(const vector<char> bitString, int startPoint, int bitLength)
{
	int i, real;

	for(i = real = 0; i < bitLength; i++)
	{
		if(bitString[startPoint + i])
			real += 1 << (bitLength - i - 1);
	}

	return real;
}


//=============================================================================
//class Gene
//=============================================================================

Gene::Gene()
{

}

Gene::~Gene()
{

}

void Gene::initialize(int geneLength)
{
	int i;

	bitString.resize(geneLength);

	for(i = 0; i < geneLength; i++)
	{
		bitString[i] = genrand_int(2);
	}
	fitness = -DBL_MAX;
}

void Gene::initialize(int geneLength, char value)
{
	int i;

	bitString.resize(geneLength);

	for(i = 0; i < geneLength; i++)
	{
		bitString[i] = value;
	}
	fitness = -DBL_MAX;
}

void Gene::initialize(int geneLength, const vector<char>& code)
{
	bitString.resize(geneLength);

	for(int i = 0; i < geneLength; i++)
	{
		bitString[i] = code[i];
	}
	fitness = -DBL_MAX;
}


void Gene::visualize()
{
	unsigned int i;

	for(i = 0; i < bitString.size(); i++)
		cerr << static_cast<int>(bitString[i]);
	cerr << endl;
}

bool Gene::operator<(const Gene& gene) const
{
	return fitness > gene.fitness;
}

//=============================================================================
//class Island
//=============================================================================

Island::Island()
{

}

Island::~Island()
{

}

void Island::initialize(int id, int numEnv, ifstream& fin)
{
        unsigned int i;
	islandId	= id;
	numIsland	= numEnv;

	gene.resize(populationSize);
	migrant.resize(migrantSize);
	elite.resize(numElite);
	newGeneration.resize(populationSize);
	crossoverTable.resize(geneLength - 1);

	for(i = 0; i < gene.size(); i++)
		gene[i].initialize(geneLength);
	if(fin.is_open())
		setGeneFromFile(fin);
	for(i = 0; i < gene.size(); i++)
		fitness(gene[i]);

	for(i = 0; i < migrant.size(); i++)
		migrant[i].initialize(geneLength);

	for(i = 0; i < elite.size(); i++)
		elite[i].initialize(geneLength);

	sort(gene.begin(), gene.end());
	copy(gene.begin(), gene.begin() + elite.size(), elite.begin());

	for(i = 0; i < newGeneration.size(); i++)
		newGeneration[i].initialize(geneLength);

	for(int i = 0; i < geneLength - 1; i++)
		crossoverTable[i] = i + 1;
}

void Island::immigrate()
{
	if(migrantSize == 0 || numIsland < 2)
		return;

	gene.resize(nativeSize);

	for(unsigned int i = 0; i < migrant.size(); i++)
		gene.push_back(migrant[i]);

	random_shuffle(gene.begin(), gene.end(), genrand_int);
}

void Island::crossover()
{
	nPointCrossover();
}

void Island::nPointCrossover()
{
	int i, j, k;
	int median;
	vector<Gene> child(2);
	vector<int> crossoverPoint;
	median = gene.size() / 2;

	for(i = 0; i < median; i++)
	{
		if(genrand() < crossoverRate)
		{
			random_shuffle(crossoverTable.begin(), crossoverTable.end(), genrand_int);

			crossoverPoint.resize(numCrossoverPoint);

			for(k = 0; k < numCrossoverPoint; k++)
				crossoverPoint[k] = crossoverTable[k];

			sort(crossoverPoint.begin(), crossoverPoint.end());

			child[0] = gene[i];
			child[1] = gene[i + median];

			for(k = 0; k < crossoverPoint.size(); k++)
			{
				for(j = crossoverPoint[k]; j < geneLength; j++)
				{
					if(k % 2 == 0)
					{
						child[0].bitString[j] = gene[i + median].bitString[j];
						child[1].bitString[j] = gene[i].bitString[j];
					}
					else
					{
						child[0].bitString[j] = gene[i].bitString[j];
						child[1].bitString[j] = gene[i + median].bitString[j];
					}
				}
			}
			gene[i] = child[0];
			gene[i + median] = child[1];
		}
	}
}


void Island::mutate()
{
	int i, j;
	double probability;

	for(i = 0; i < gene.size(); i++)
	{
		for(j = 0; j < gene[i].bitString.size(); j++)
		{
			probability = genrand();
			if(probability < mutationRate)
				gene[i].bitString[j] ^= 1;
		}
	}
}

void Island::evaluate()
//̂̕]
{
	int i;

	for(i = 0; i < gene.size(); i++)
		fitness(gene[i]);
}


void Island::elitePreservation()
// G[g̕ۑ
//O̐̃G[gƌ݂̐̑S̂킹āC̏ʂG[gƂD
//WcTCY͈ꎞIpopSize + numElite ɂȂD
{
	int i;
	vector<Gene> temp;
	temp = elite;

	sort(gene.begin(), gene.end());

	for(i = 0; i < elite.size(); i++)
		temp.push_back(gene[i]);

	sort(temp.begin(), temp.end());

	for(i = 0; i < elite.size(); i++){
		// G[g̂XV
		elite[i] = temp[i];
		// G[g߂
		gene.push_back(temp[i]);
	}

	bestFitness = elite[0].fitness;
}


void Island::setGeneFromFile(ifstream& fin)
// e̂̐F̃rbgCt@Cǂݍ񂾐݌vϐlɐݒ肷D
// ݌vϐ -> R[h -> e̂̈`qɊi[D
// t@C̈s1́Dŏ̌̂珇ɓǂݍ
{
	for(int i = 0; i < gene.size(); ++i){
		vector<char>& bitString = gene[i].bitString;
		for(int j = 0; j < numVariable; ++j){
			double variable;
			fin >> variable;

			if(fin.fail()) // ǂݍݎsŏI
				return;

			// Ɉ߂
			if(variable < closedMin)
				variable = closedMin;
			else if(variable >= openMax)
				variable = openMax - delta;

			// l琮lɕϊ (ۂߌ덷͎ľܓ)
			int num = static_cast<int>((variable - closedMin) / delta + 0.5);

			// R[hgene[i].bitString[] Ɋi[
			gray_code(bitString, j * variableLength, variableLength, num);
		}
	}
}


void Island::reproduce()
{
	if(selectionType == 0)
		tournamentSelection();
	else if(selectionType == 1)
		rouletteSelection();
	else{
		cerr << endl << "Invalid selection" << endl << "Exit" << endl;
		exit(1);
	}

	//̂̒u
	//genexN^̑傫WcTCYɖ߂D
	gene.resize(newGeneration.size());
	for(int i = 0; i < gene.size(); i++)
		gene[i] = newGeneration[i];
}


//g[igI
//p[^FtournamentSize
void Island::tournamentSelection()
{
	int i,j;

	vector<Gene> tournament;
	tournament.resize(tournamentSize);

	for(i=0; i< newGeneration.size(); i++){
		for(j=0; j< tournament.size(); j++){
			tournament[j] = gene[genrand_int(gene.size())];
		}
		sort(tournament.begin(),tournament.end());
		newGeneration[i] = tournament[0];
	}
}

// [bgI
void Island::rouletteSelection()
{
	// [bgTsum_of_fitness߂D
	// ̂̓Kx̑a
	double min_fit = DBL_MAX;
	for(unsigned i = 0; i < gene.size(); ++i)
		if(min_fit > gene[i].fitness)
			min_fit = gene[i].fitness;

	// make roulette
	roulette.resize(gene.size());
	roulette[0] = gene[0].fitness - min_fit;
	for(unsigned i = 1; i < gene.size(); ++i)
		roulette[i] = roulette[i-1] + gene[i].fitness - min_fit;
	double sum_of_fitness = roulette.back();
	assert(sum_of_fitness > 0);

	// [bgnew_size 񂾂񂵂Ď̌̂I
	for(unsigned i = 0; i < newGeneration.size(); i++){
		// [bg̑Iʒu(0 <= r <= sum_of_fitness)
		double r = sum_of_fitness * genrand();

		vector<double>::iterator itr =
			std::lower_bound(roulette.begin(), roulette.end(), r);

		// `q̃Rs[
		newGeneration[i] = gene[itr - roulette.begin()];
	}
}


void Island::fitness(const Gene& gene) const
{
	int		i, j;
	double	d, e;
	vector<double> real(numVariable);

	for(i = 0; i < numVariable; i++){
		if(typeOfEncode == 0)
			real[i] = gray_decode(gene.bitString, i * variableLength, variableLength) * delta + closedMin;
		else
			real[i] = binary_decode(gene.bitString, i * variableLength, variableLength) * delta + closedMin;
	}

	if(functionType == 0) //rastrigin
	{
		for(i = 0, d = 0; i < numVariable; i++)
			d += real[i] * real[i] - 10 * cos(2 * 3.14159265359 * real[i]);
		gene.fitness = -(10 * numVariable + d);
	}
	else if(functionType == 1) //schwefel
	{
		for(i = 0, d = 0; i < numVariable; i++)
			d += -1 * real[i] * sin(sqrt(fabs(real[i])));
		gene.fitness = -(d + (418.98288 * 10) - 0.00115983);
	}
	else if(functionType == 2) //griewank
	{
		for(i = 0, d = 1, e = 1; i < numVariable; i++)
		{
			d += real[i] * real[i] / 4000;
			e *= cos(real[i] / sqrt((double)(i + 1)));
		}
		gene.fitness = -(d - e);
	}
	else if(functionType == 3) //rosenbrock
	{
		for(i = 1, d = 0; i < numVariable; i++)
			d += 100 * (real[0] - real[i] * real[i]) * (real[0] - real[i] * real[i])
			+ (real[i] - 1) * (real[i] - 1);
		gene.fitness = -(d);
	}
	else if(functionType == 4) //ridge
	{
		for(i = 0, d = 0; i < numVariable; i++)
		{
			for(j = 0, e = 0; j <= i; j++)
				e += real[j];
			d += e * e;
		}
		gene.fitness = -(d);
	}
	else if(functionType == 5) //onemax
	{
		int bit1 = accumulate(gene.bitString.begin(), gene.bitString.end(), 0);
		gene.fitness = bit1 - static_cast<int>(gene.bitString.size());
	}
	else{
		cerr << endl << "invalid function" << endl << "Exit" << endl;
		exit(1);
	}

	numEvaluate++;

// debug
//	copy(real.begin(), real.end(), ostream_iterator<double>(cerr, "\t"));
//	cerr << " \t" << gene.fitness << endl;
}

//=============================================================================
//class Globe
//=============================================================================

Globe::Globe()
{

}

Globe::~Globe()
{

}

void Globe::initialize()
{
// Default settings

	fTotalPopSize = false;
	totalPopulationSize			= 400;

	island_0.islandId			= 0;

	// parameters for chromosome
	island_0.variableLength		= 10;
	island_0.numVariable		= 10;
	island_0.geneLength			= island_0.numVariable * island_0.variableLength;

	fInitialPopulation			= 0; // 0: random,  1: from file(design variables)

	// parameters for objective function
	island_0.functionType		= 0;
	island_0.openMax			= 5.12;
	island_0.closedMin			= -5.12;
	island_0.delta				= 10.24  / ((double) (1 << island_0.variableLength)); //0.01;
	island_0.finValue			= -0.00000001;

	// canonical GA parameters
	island_0.populationSize  	= 10; // in each subpopulations
	island_0.crossoverRate		= 1.0;
	island_0.mutationRate		= 0.01;
	island_0.numElite			= 5;
	island_0.tournamentSize     = 4;
	island_0.selectionType		= 0; // 0: tournament,  1: roulette

	// DGA parameters
	island_0.migrationRate		= 0.5;
	numIsland					= 40;
	migrationInterval			= 5;


	island_0.numCrossoverPoint	= 2;
	island_0.typeOfEncode		= 0;


	island_0.numEvaluate		= 0;
	maxGeneration		  		= 1000;


	maxTrial			= 1;
	sumMaxFitness		= 0;
	sumGeneration		= 0;
	numConvergent		= 0;
	sumTotalEvaluate	= 0;
	fStop				= true;
	fOutput				= false;
}

void Globe::initialize(int argc, char *argv[])
{
	initialize();

	// command line
	for(int i = 1; i < argc; i++)
	{
		const char *p = argv[i] + 1;
		switch(*argv[i])
		{
		case 'b':
			island_0.geneLength = atoi(p);
			island_0.numVariable = island_0.geneLength / island_0.variableLength;
			break;

		case 'i':
			numIsland = atoi(p);
			break;

		case 'c':
			island_0.crossoverRate = atof(p);
			break;

		case 'f':
			island_0.functionType = atoi(p);
			if(atoi(p) == 0){
				island_0.closedMin = -5.12;
				island_0.openMax = 5.12;
				island_0.delta = 0.01;
			}
			else if(atoi(p) == 1 || atoi(p) == 2){
				island_0.closedMin = -512;
				island_0.openMax = 512;
				island_0.delta = 1;
			}
			else if(atoi(p) == 3){
				island_0.closedMin = -2.048;
				island_0.openMax = 2.048;
				island_0.delta = 0.004;
			}
			else if(atoi(p) == 4){
				island_0.closedMin = -64;
				island_0.openMax = 64;
				island_0.delta = 0.125;
			}
			break;

		case 'I':
			migrationInterval = atoi(p);
			break;

		case 'm':
			island_0.mutationRate = atof(p);
			break;

		case 'M':
			island_0.migrationRate = atof(p);
			break;

		case 'o':
			fOutput = true;
			break;

		case 'p':
			island_0.populationSize = atoi(p);
			island_0.numElite = island_0.populationSize / 2;
			break;

		case 'P':
			totalPopulationSize = atoi(p);
			fTotalPopSize = true;
			break;

		case 'z':
			maxTrial = atoi(p);
			break;

		case 'g':
			maxGeneration = atoi(p);
			break;

		case 'S':
			// 0: random,  1: from file(design variables)
			fInitialPopulation	= atoi(p);
			break;

		case 's':
			// 0: tournament,  1: roulette
			island_0.selectionType = atoi(p);
			break;
		}
	}
}

void Globe::establish()
{
	if(fTotalPopSize)
	{
		island_0.populationSize = 10;
		island_0.numElite = island_0.populationSize / 2;
		numIsland = totalPopulationSize / island_0.populationSize;
	}
	island_0.migrantSize	= static_cast<int>(island_0.populationSize * island_0.migrationRate);
	island_0.nativeSize	= island_0.populationSize - island_0.migrantSize;

	populationSize	= island_0.populationSize;
	nativeSize		= island_0.nativeSize;
	migrantSize		= island_0.migrantSize;

	maxFitness		= -DBL_MAX;
	convGeneration	= maxGeneration;
	fConv			= false;
	totalEvaluate	= 0;

	island.resize(numIsland);

	int i;

	ifstream fin;
	if(fInitialPopulation == 1){
		// ݌vϐl̃t@CI[v
		fin.open(INITIAL_POPULATION_FILE);
		if(!fin.is_open()){
			cerr << __FILE__ << ":(" << __LINE__ << ") can't open file "
				<< INITIAL_POPULATION_FILE << endl;
			exit(1);
		}
	}
	for(i = 0; i < island.size(); i++)
	{
		island[i] = island_0;
		island[i].initialize(i, island.size(), fin);
	}

	fin.close();
}

void Globe::emigrate()
{
	if(migrantSize == 0 || island.size() < 2)
		return;

	int i, j;
	RandomRing	destination(numIsland);

	// siڏZg|W̓_Oj
	destination.shuffle();

	for(i = 0; i < island.size(); i++)
	{
		// ڏZ̂̓_
		random_shuffle( island[i].gene.begin(), island[i].gene.end(), genrand_int );
		// s̓ migrant Ɍ̂Rs[
		for(j = 0; j < migrantSize; j++)
			island[ destination[i] ].migrant[j] = island[i].gene[nativeSize + j];
	}
}


void Globe::execute()
{
	int i;
	int id = 0;

	for(generation = 0; generation < maxGeneration; generation++)
	{
	  //if(generation % 10 == 0)
	  //	cerr << "." << flush;

		if(generation % migrationInterval == 0)
			emigrate();

		for(i = 0; i < island.size(); i++)
		{
			if(generation % migrationInterval == 0)
				island[i].immigrate();

			island[i].crossover();
			island[i].mutate();
			island[i].evaluate();
			island[i].elitePreservation();

			if(island[i].bestFitness > maxFitness)
				maxFitness = island[id=i].bestFitness;

		}

		cout << maxFitness <<  " " ;
		for (int hoge = 0; hoge < island[id].gene[0].bitString.size(); ++hoge)
		  cout << (char)((int)island[id].gene[1].bitString[hoge] + (int)'0') ;
		cout << endl;

		for(i = 0; i < island.size(); i++)
			island[i].reproduce();

		if(!fConv && maxFitness >= island_0.finValue)
			finalize(true);

		sumFitness[generation] += maxFitness;
	}
	finalize(false);
}

void Globe::finalize(bool b)
{
	int i;

	if(b)
	{
		maxFitness = 0.000000;
		sumMaxFitness += maxFitness;
		convGeneration = generation;
		sumGeneration += convGeneration;
		numConvergent++;
		cerr << "*" << flush;
		fConv = true;

		if(fStop)
			for(; generation < maxGeneration; generation++)
				sumFitness[generation] += maxFitness;

		for(i = 0; i < island.size(); i++)
			totalEvaluate += island[i].numEvaluate;
		sumTotalEvaluate += totalEvaluate;
	}
	else if(!b && !fConv)
	{
		sumMaxFitness += maxFitness;
		for(i = 0; i < island.size(); i++)
			totalEvaluate += island[i].numEvaluate;
		sumTotalEvaluate += totalEvaluate;
	}
}


//sOɃp[^cerrŏo͂
void Globe::showParameters()
{
	cerr << "\nPopulation     : " << island_0.populationSize * numIsland << endl;
	cerr << "Num island     : " << numIsland << endl;
	cerr << "Pop/island     : " << island_0.populationSize << endl;
	cerr << "init population: " << ((fInitialPopulation == 0) ? "random" : "from file") << endl;
	cerr << "Elite/island   : " << island_0.numElite << endl;
	cerr << "MigRate        : " << island_0.migrationRate << endl;
	cerr << "MigInterval    : " << migrationInterval << endl ;
	cerr << "Pc             : " << island_0.crossoverRate << endl;
	cerr << "Pm             : " << island_0.mutationRate << endl;;
	cerr << "MaxGeneration  : " << maxGeneration << endl;

	//֐Ɋւ
	cerr << "Function       : " <<  "f" << island_0.functionType << endl ;
	cerr << "\n";

}


